﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("CrossoverPerformanceAnalyzer", "An operator that analyzes the performance of crossovers.")]
  [StorableClass]
  public class CrossoverPerformanceAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string GenerationsParameterName = "Generations";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion


    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public ILookupParameter<ItemArray<Permutation>> ParentsParameter {
      get { return (ScopeTreeLookupParameter<Permutation>)Parameters["Parents"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> ParentsQualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["ParentsQuality"]; }
    }
    public ILookupParameter<Permutation> ChildParameter {
      get { return ((LookupParameter<Permutation>)Parameters["Child"]); }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public IValueParameter<SingleObjectiveSolutionSimilarityCalculator> SimilarityCalculatorParameter {
      get { return (IValueParameter<SingleObjectiveSolutionSimilarityCalculator>)Parameters["SimilarityCalculator"]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    #endregion

    ScatterPlot plot, childDiversityPlot, parentDiversityPlot;
    DataRow dtRow, dtRowSuccess;
    int cnt = 0;
    int success = 0;

    [StorableConstructor]
    private CrossoverPerformanceAnalyzer(bool deserializing) : base(deserializing) { }
    private CrossoverPerformanceAnalyzer(CrossoverPerformanceAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public CrossoverPerformanceAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));

      Parameters.Add(new ScopeTreeLookupParameter<Permutation>("Parents", "The parent permutations which have been crossed."));
      ParentsParameter.ActualName = "TSPTour";

      Parameters.Add(new LookupParameter<Permutation>("Child", "The child permutation resulting from the crossover."));
      ChildParameter.ActualName = "TSPTour";

      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("ParentsQuality", "The quality of the parent solutions."));
      ParentsQualityParameter.ActualName = "TSPTourLength";

      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The evaluated quality of the child solution."));
      QualityParameter.ActualName = "TSPTourLength";

      Parameters.Add(new ValueParameter<SingleObjectiveSolutionSimilarityCalculator>("SimilarityCalculator"));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CrossoverPerformanceAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      SimilarityCalculatorParameter.Value.QualityVariableName = "TSPTourLength";
      SimilarityCalculatorParameter.Value.SolutionVariableName = "TSPTour";

      Point2D<double> qualityPoint, diversityPointChild, diversityPointParent;
      var qualityParent1 = ParentsQualityParameter.ActualValue.First().Value;
      var qualityParent2 = ParentsQualityParameter.ActualValue.Last().Value;
      var child = ChildParameter.ActualValue;
      var parent1 = ParentsParameter.ActualValue.First();
      var parent2 = ParentsParameter.ActualValue.Last();
      var parentDiversity = SimilarityCalculatorParameter.Value.CalculateSolutionSimilarity(ExecutionContext.Scope.SubScopes.First(), ExecutionContext.Scope.SubScopes.Last());

      diversityPointParent = new Point2D<double>(cnt, parentDiversity);
      double worseQuality = qualityParent1 > qualityParent2 ? qualityParent1 : qualityParent2;
      if (qualityParent1 > qualityParent2) {
        diversityPointChild = new Point2D<double>(cnt, SimilarityCalculatorParameter.Value.CalculateSolutionSimilarity(ExecutionContext.Scope, ExecutionContext.Scope.SubScopes.First()));
      } else {
        diversityPointChild = new Point2D<double>(cnt, SimilarityCalculatorParameter.Value.CalculateSolutionSimilarity(ExecutionContext.Scope, ExecutionContext.Scope.SubScopes.Last()));
      }

      qualityPoint = new Point2D<double>(cnt++, worseQuality - QualityParameter.ActualValue.Value);
      if ((worseQuality - QualityParameter.ActualValue.Value) > 0) {
        success++;
      }

      string curGenStr = GenerationsParameter.ActualValue.Value.ToString();

      if (!Results.ContainsKey("Scatter Plot")) {
        InitializePlot();
        InitializeChildDiversityPlot();
        InitializeParentDiversityPlot();

        Results.Add(new Result("Scatter Plot", plot));
        Results.Add(new Result("Scatter Plot History", new ScatterPlotHistory()));

        Results.Add(new Result("Child Diversity Scatter Plot", childDiversityPlot));
        Results.Add(new Result("Child Diversity Scatter Plot History", new ScatterPlotHistory()));

        Results.Add(new Result("Parent Diversity Scatter Plot", parentDiversityPlot));
        Results.Add(new Result("Parent Diversity Scatter Plot History", new ScatterPlotHistory()));
        cnt = 0;

        DataTable dt = new DataTable("Average Crossover Performance");
        dtRow = new DataRow("Average Crossover Performance per Generation");
        dt.Rows.Add(dtRow);
        Results.Add(new Result("Average Crossover Performance", dt));

        DataTable dtSuccess = new DataTable("Successfull Crossovers");
        dtRowSuccess = new DataRow("Successfull Crossovers per Generation");
        dtSuccess.Rows.Add(dtRowSuccess);
        Results.Add(new Result("Successfull Crossovers", dtSuccess));

        success = 0;
      }

      if (!plot.Rows.ContainsKey(curGenStr)) {
        if (GenerationsParameter.ActualValue.Value != 0) {
          double avg = plot.Rows[(GenerationsParameter.ActualValue.Value - 1).ToString()].Points.Average(x => x.Y);
          dtRow.Values.Add(avg);
          dtRowSuccess.Values.Add(success);

          ((ScatterPlotHistory)Results["Scatter Plot History"].Value).Add(plot);
          ((ScatterPlotHistory)Results["Child Diversity Scatter Plot History"].Value).Add(childDiversityPlot);
          ((ScatterPlotHistory)Results["Parent Diversity Scatter Plot History"].Value).Add(parentDiversityPlot);
          InitializePlot();
          InitializeChildDiversityPlot();
          InitializeParentDiversityPlot();
          Results["Scatter Plot"].Value = plot;
          Results["Child Diversity Scatter Plot"].Value = childDiversityPlot;
          Results["Parent Diversity Scatter Plot"].Value = parentDiversityPlot;
          cnt = 0;
          success = 0;
        }

        var points = new List<Point2D<double>>();
        points.Add(qualityPoint);
        ScatterPlotDataRow row;
        row = new ScatterPlotDataRow(curGenStr, null, points);
        row.VisualProperties.PointStyle = ScatterPlotDataRowVisualProperties.ScatterPlotDataRowPointStyle.Circle;
        row.VisualProperties.PointSize = 5;
        plot.Rows.Add(row);

        points = new List<Point2D<double>>();
        points.Add(diversityPointChild);
        row = new ScatterPlotDataRow(curGenStr, null, points);
        row.VisualProperties.PointStyle = ScatterPlotDataRowVisualProperties.ScatterPlotDataRowPointStyle.Circle;
        row.VisualProperties.PointSize = 5;
        childDiversityPlot.Rows.Add(row);

        points = new List<Point2D<double>>();
        points.Add(diversityPointParent);
        row = new ScatterPlotDataRow(curGenStr, null, points);
        row.VisualProperties.PointStyle = ScatterPlotDataRowVisualProperties.ScatterPlotDataRowPointStyle.Circle;
        row.VisualProperties.PointSize = 5;
        parentDiversityPlot.Rows.Add(row);
      } else {
        plot.Rows[curGenStr].Points.Add(qualityPoint);
        childDiversityPlot.Rows[curGenStr].Points.Add(diversityPointChild);
        parentDiversityPlot.Rows[curGenStr].Points.Add(diversityPointParent);
      }

      return base.Apply();
    }

    private void InitializePlot() {
      plot = new ScatterPlot("Crossover Performance", null);
      plot.VisualProperties.XAxisTitle = "Solution Index";
      plot.VisualProperties.YAxisTitle = "Absolut Quality Difference";
    }
    private void InitializeChildDiversityPlot() {
      childDiversityPlot = new ScatterPlot("Child Diversity", null);
      childDiversityPlot.VisualProperties.XAxisTitle = "Solution Index";
      childDiversityPlot.VisualProperties.YAxisTitle = "Diversity";
    }
    private void InitializeParentDiversityPlot() {
      parentDiversityPlot = new ScatterPlot("Parent Diversity", null);
      parentDiversityPlot.VisualProperties.XAxisTitle = "Solution Index";
      parentDiversityPlot.VisualProperties.YAxisTitle = "Diversity";
    }
  }
}
